/*	
 *	flash.c
 *
 *	P/EMU - P/ECE Emulator
 *	Copyright (C) 2003 Naoyuki Sawa
 *
 *	* Mon Apr 14 00:00:00 JST 2003 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

#define SECTOR_SIZE	0x01000		/* ZN^  P= 4KB */
#define BLOCK_SIZE	0x10000		/* ubNP=64KB */

/****************************************************************************
 *	O[oϐ
 ****************************************************************************/

FLASH flash;

/****************************************************************************
 *	O[o֐
 ****************************************************************************/

void
flash_init()
{
#define FLASH_TOP	0x0c00000
#define READ_UNIT	0x1000

	int retval, size, bits, ofs;

	memset(&flash, 0, sizeof flash);

	/* CFING[쐬B(Ƃ肠KvȃtB[h) */
	size = (int)emu.sysinfo.pffs_end - FLASH_TOP;
	bits = 0;
	for(;;) {
		if(1 << bits >= size) break;
		bits++;
	}
	flash.cfiinfo.device_size = bits;

	/* 蓖āB */
	flash.mem_size = 1 << flash.cfiinfo.device_size;
	flash.mem = (unsigned char*)calloc(flash.mem_size, 1);

	/* C[Wǂݍ݁B */
	printf("tbVǂݍݒ");
	for(ofs = 0; ofs < flash.mem_size; ofs += READ_UNIT) {
		retval = ismReadMem(&flash.mem[ofs], FLASH_TOP + ofs, READ_UNIT);
		if(retval) DIE();
		if((ofs / READ_UNIT) % 10 == 9) printf(".");
	}
	printf("OK\n");

#undef FLASH_TOP
#undef READ_UNIT
}

int
flash_read(unsigned ofs, int size)
{
	int data = 0;

	ofs &= flash.mem_size - 1;

	switch(flash.stat) {
	case FLASH_NORMAL:
		memcpy(&data, &flash.mem[ofs], size);
		break;
	case FLASH_CFI_QUERY:
		/* \tgEFAID[hΉ */
		ofs -= CFIINFO_OFFSET * sizeof(short);
		if(ofs > sizeof(CFIINFO)) DIE();
		memcpy(&data, (char*)&flash.cfiinfo + ofs, size);
		break;
	default:
		DIE();
	}

	return data;
}

void
flash_write(unsigned ofs, int data, int size)
{
	int cmd;

	if(size != 2) DIE(); /* ݂̓n[t[hPʂ̂݁I */
	ofs &= flash.mem_size - 1;
	cmd = data & 0xff;

	/* WORD PROGRAMȊOł0xf0݂́AėpEXITR}hƌȂ܂B */
	if(flash.stat != FLASH_WORD_PROGRAM && cmd == 0xf0) {
		flash.stat = FLASH_NORMAL; /* MANUAL EXIT */
		return;
	}

	switch(flash.stat) {
	case FLASH_NORMAL:
		if(ofs == 0x5555 * 2 && cmd == 0xaa) {
			flash.stat = FLASH_ENTER1;
		} else {
			DIE();
		}
		break;
	case FLASH_ENTER1:
		if(ofs == 0x2aaa * 2 && cmd == 0x55) {
			flash.stat = FLASH_ENTER2;
		} else {
			DIE();
		}
		break;
	case FLASH_ENTER2:
		if(ofs == 0x5555 * 2) {
			switch(cmd) {
			case 0x80:
				flash.stat = FLASH_ENTER3;
				break;
			case 0x90: /* SOFTWARE ID                      */
			case 0x98: /* CFI QUERY   ́ÂƂĈ */
				flash.stat = FLASH_CFI_QUERY;
				break;
			case 0xa0:
				flash.stat = FLASH_WORD_PROGRAM;
				break;
			default:
				DIE();
			}
		} else {
			DIE();
		}
		break;
	case FLASH_ENTER3:
		if(ofs == 0x5555 * 2 && cmd == 0xaa) {
			flash.stat = FLASH_ENTER4;
		} else {
			DIE();
		}
		break;
	case FLASH_ENTER4:
		if(ofs == 0x2aaa * 2 && cmd == 0x55) {
			flash.stat = FLASH_ENTER5;
		} else {
			DIE();
		}
		break;
	case FLASH_ENTER5:
		switch(cmd) {
		case 0x10: /* CHIP ERASE */
			if(ofs == 0x5555 * 2) {
				dbg("CHIP ERASE");
				memset(flash.mem, -1, flash.mem_size);
				flash.stat = FLASH_NORMAL; /* AUTO EXIT */
			} else {
				DIE();
			}
			break;
		case 0x30: /* SECTOR ERASE */
			ofs &= ~(SECTOR_SIZE - 1);
			dbg("SECTOR ERASE: %07x", ofs);
			memset(&flash.mem[ofs], -1, SECTOR_SIZE);
			flash.stat = FLASH_NORMAL; /* AUTO EXIT */
			break;
		case 0x50: /* BLOCK ERASE */
			ofs &= ~(BLOCK_SIZE - 1);
			dbg("BLOCK ERASE: %07x", ofs);
			memset(&flash.mem[ofs], -1, BLOCK_SIZE);
			flash.stat = FLASH_NORMAL; /* AUTO EXIT */
			break;
		default:
			DIE();
		}
		break;
	case FLASH_CFI_QUERY:
		if((ofs == 0x5555 * 2 && cmd == 0xaa) ||
		   (ofs == 0x2aaa * 2 && cmd == 0x55)) {
			/* EXITV[PXO(aa-55-f0)̂As񖽗߂𖳎܂B
			 * Ō̈ꖽ߂͔ėpEXITR}hƂāÅ֐̍ŏł܂Ƃ߂ďĂ܂B
			 */
		} else {
			DIE();
		}
		break;
	case FLASH_WORD_PROGRAM:
		dbg("WORD PROGRAM: %07x=%04x", ofs, data & 0xffff);
		memcpy(&flash.mem[ofs], &data, size);
		flash.stat = FLASH_NORMAL; /* AUTO EXIT */
		break;
	default:
		DIE();
	}
}

